@@ -74,7 +74,7 @@
 #  ifndef HAVE_STRRCHR
 #    define strrchr rindex
 #  endif
-char *strchr (), *strrchr ();
+   char *strchr (), *strrchr ();
 #  ifndef HAVE_MEMMOVE
 #    define memmove(d, s, n) bcopy ((s), (d), (n))
 #  endif
@@ -94,13 +94,38 @@ char *strchr (), *strrchr ();
 
 
 
-#include <readline/readline.h>
-#include <readline/history.h>
-
-
 #define BUFFSIZE 512
 
 
+#ifdef  HAVE_SNPRINTF  /* don't rely on the compiler understanding variadic macros */
+# define snprintf1(buf,bufsize,format,arg1) snprintf(buf,bufsize,format,arg1)
+#else
+# define snprintf1(buf,bufsize,format,arg1) sprintf(buf,format,arg1)
+# define vsnprintf(buf,bufsize,format,ap)   vsprintf(buf,format,ap)
+#endif
+
+
+#ifndef HAVE_STRNLEN
+# define strnlen(s,l) strlen(s)
+#endif
+
+
+
+#include <readline/readline.h>
+#include <readline/history.h>
+
+#ifdef SPY_ON_READLINE
+# ifndef HOMEGROWN_REDISPLAY
+#  define MAYBE_MULTILINE 1
+# endif
+#endif
+
+#ifdef MAYBE_MULTILINE
+extern int _rl_horizontal_scroll_mode; /* Spying on readline's private life .... */
+#  define redisplay_multiple_lines (!_rl_horizontal_scroll_mode)
+#else
+#  define redisplay_multiple_lines FALSE
+#endif
 
 /* in main.c: */
 extern int master_pty_fd; 
@@ -116,44 +141,48 @@ extern int debug;
 extern char *password_prompt_search_string;
 
 void init_rlwrap(void);
-void cleanup_rlwrap_and_exit(int);
-void main_loop(int);
+void cleanup_rlwrap_and_exit(int status);
+void main_loop(int pty_fd);
 
 
 /* in signals.c */
 extern int child_is_dead, childs_exit_status, sigterm_received;
 void install_signal_handlers(void);
-void block_signals(int*);
-void unblock_signals(int*);
+void block_signals(int* sigs);
+void unblock_signals(int* sigs);
 void ignore_sigchld(void);
 
 /* in utils.c */
 void usage(void);
-void mywarn(const char *, ...);
-void myerror(const char *, ...);
-void *mymalloc(size_t);
-char *mybasename(char *);
-char *mystrndup(const char *, int);
-char *mysavestring(const char*);
-char *add3strings(const char*, const char*, const char*);
+void mywarn(const char  *message, ...);
+void myerror(const char *message, ...);
+void *mymalloc(size_t size);
+char *mybasename(char *filename);
+char *mydirname(char *filename);
+char *mystrndup(const char *string, int len);
+void mystrlcpy(char *dst, const char *src, size_t size);
+void mystrlcat(char *dst, const char *src, size_t size);
+char *mysavestring(const char*string);
+char *add3strings(const char *str1, const char *str2, const char *str3);
 #define add2strings(a,b)  add3strings(a,b,"")
-char *mangle_string_for_debug_log(char *, int);
-void open_logfile(const char *);
-void write_logfile(const char *);
+char *mangle_string_for_debug_log(char *string, int maxlen);
+void open_logfile(const char *filename);
+void write_logfile(const char *str);
 void close_logfile(void);
 
 /* in readline.c: */
 extern struct rl_state {      /* struct to save readline text buffer and point */
   char text[BUFFSIZE];
   char prompt[BUFFSIZE];
-  int point;
-  int already_saved;         /* flag set when saved, cleared when restored */ 
+  int  point;
+  int  already_saved;         /* flag set when saved, cleared when restored */ 
 } saved_rl_state;
 
 void save_rl_state (struct rl_state *);
 void restore_rl_state (struct rl_state *);
 void init_readline(char *);
 void line_handler(char *);
+void my_redisplay(void);
 
 extern  int within_line_edit, transparent;
 extern  char previous_line[];
@@ -176,7 +205,7 @@ void init_completer(const char *);
 void feed_into_completion_list(const char *);
 char *my_completion_function(char *, int);
 
-extern int  remember_for_completion, use_completion_list;
+extern int  remember_for_completion;
 
 /* in term.c */
 int redisplay;                         /* TRUE when user input should be visible */  
@@ -186,12 +215,15 @@ void prepare_terminal(void);
 void cr(void);
 void backspace(int);
 void clear_line(void);
-int  my_putchar(int);
-
+void clear_the_screen(void);
+void curs_up(void);
+void curs_down(void);
+int  my_putchar(int c);
+void cursor_hpos (int col);
 extern struct termios saved_terminal_settings;
 extern int terminal_settings_saved;
 extern struct winsize window_size;
-extern char *term_backspace, term_eof, term_stop;
+extern char *term_backspace, term_eof, term_stop, *term_cursor_hpos, *term_cursor_up, *term_cursor_down;
 
 /* some handy macros */
 #ifndef TRUE
@@ -202,10 +234,13 @@ extern char *term_backspace, term_eof, term_stop;
 #  define FALSE 0
 #endif
 
+#ifndef min
+# define min(a,b) ((a) < (b) ? (a) : (b))
+#endif
 
-
-
-
+#ifndef max
+# define max(a,b) ((a) < (b) ? (b) : (a))
+#endif
 
 
 
@@ -238,6 +273,7 @@ extern char *term_backspace, term_eof, term_stop;
 #  define DPRINTF2(mask, format,arg1, arg2)
 #  define DPRINTF3(mask, format,arg1, arg2, arg3)
 #  define ERRMSG(b) 
+#  define NDEBUG    /* disable assertions */
 #endif
 
 
